unit mhotkeyedit;

interface

uses
  Buttons,
  Classes,
  ComCtrls,
  Controls,
  ExtCtrls,
  Graphics,
  Math,
  Menus,
  StdCtrls,
  SysUtils,
  Themes,
  Windows;

type
  TSCHKModifier = (schkShift, schkCtrl, schkAlt, schkExt);
  TSCHKModifiers = set of TSCHKModifier;
  TSCHKInvalidKey = (schcNone, schcShift, schcCtrl, schcAlt,
    schcShiftCtrl, schcShiftAlt, schcCtrlAlt, schcShiftCtrlAlt);
  TSCHKInvalidKeys = set of TSCHKInvalidKey;

const
  scDefaultModifiers = [schkAlt];
  scDefaultInvalidKeys = [schcNone, schcShift];

const
  scskNone = '';
  scskBkSp = 'BkSp';
  scskTab = 'Tab';
  scskEsc = 'Esc';
  scskEnter = 'Enter';
  scskSpace = 'Space';
  scskPgUp = 'PgUp';
  scskPgDn = 'PgDn';
  scskEnd = 'End';
  scskHome = 'Home';
  scskLeft = 'Left';
  scskUp = 'Up';
  scskRight = 'Right';
  scskDown = 'Down';
  scskIns = 'Ins';
  scskDel = 'Del';
  scskShift = 'Shift+';
  scskCtrl = 'Ctrl+';
  scskAlt = 'Alt+';

    {$EXTERNALSYM SC_LBUTTON}
  SC_LBUTTON = 1;
    {$EXTERNALSYM SC_RBUTTON}
  SC_RBUTTON = 2;
    {$EXTERNALSYM SC_SHIFT}
  SC_SHIFT = 4;
    {$EXTERNALSYM SC_CONTROL}
  SC_CONTROL = 8;
    {$EXTERNALSYM SC_MBUTTON}
  SC_MBUTTON = 16;
    {$EXTERNALSYM SC_MENU}
  SC_MENU = 32;

type
  TSCKeyCap = (sckcBkSp, sckcTab, sckcEsc, sckcEnter, sckcSpace, sckcPgUp,
    sckcPgDn, sckcEnd, sckcHome, sckcLeft, sckcUp, sckcRight, sckcDown, sckcIns,
    sckcDel, sckcShift, sckcCtrl, sckcAlt);

var
  SCKeyCaps: array[TSCKeyCap] of string = (scskBkSp, scskTab, scskEsc, scskEnter, scskSpace, scskPgUp, scskPgDn, scskEnd, scskHome, scskLeft, scskUp, scskRight,
    scskDown, scskIns, scskDel, scskShift, scskCtrl, scskAlt);

{ THotkeyEdit }

type
  THotkeyEdit = class(TCustomEdit)
  private
    FInternalUpdate: Integer;
    FHotkeyUpdate: Integer;
    FHotKey: Word;
    FModifiers: TSCHKModifiers;
    FHotModifiers: TSCHKModifiers;
    FInvalidKeys: TSCHKInvalidKeys;

    procedure UpdateHotKey(Value: TShortCut);
    function GetHotKey: TShortCut;
    procedure SetHotKey(Value: TShortCut);
    procedure ShortCutToHotKey(Value: TShortCut);
    procedure UpdateHotText;
  protected
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure RealSetText(const AValue: TCaption); override;
    procedure KeyPress(var Key: char); override;
    procedure KeyUp(var Key: Word; Shift: TShiftState); override;
  public
    property HotKey: TShortCut read GetHotKey write SetHotKey;
  published
    property Align;
    property OnChange;
  end;

function ShortCutToText(ShortCut: TShortCut): string;

implementation

function scShortCut(Key: Word; Shift: TShiftState): TShortCut;
begin
  Result := 0;
  if WordRec(Key).Hi <> 0 then
    Exit;
  Result := Key;
  if ssShift in Shift then
    Inc(Result, scShift);
  if ssCtrl in Shift then
    Inc(Result, scCtrl);
  if ssAlt in Shift then
    Inc(Result, scAlt);
end;

function scGetSpecialKeyName(ShortCut: TShortCut): string;
var
  ScanCode: Integer;
  KeyName: array[0..255] of Char;
begin
  Result := '';
  ScanCode := MapVirtualKey(WordRec(ShortCut).Lo, 0) shl 16;
  if ScanCode <> 0 then
  begin
    GetKeyNameText(ScanCode, KeyName, SizeOf(KeyName));
    Result := KeyName;
  end;
end;

function ShortCutToText(ShortCut: TShortCut): string;
var
  Name: string;
begin
  case WordRec(ShortCut).Lo of
    $08, $09:
      Name := SCKeyCaps[TSCKeyCap(Ord(sckcBkSp) + WordRec(ShortCut).Lo - $08)];
    $0D: Name := SCKeyCaps[sckcEnter];
    $1B: Name := SCKeyCaps[sckcEsc];
    $20..$28:
      Name := SCKeyCaps[TSCKeyCap(Ord(sckcSpace) + WordRec(ShortCut).Lo - $20)];
    $2D..$2E:
      Name := SCKeyCaps[TSCKeyCap(Ord(sckcIns) + WordRec(ShortCut).Lo - $2D)];
    $30..$39: Name := Chr(WordRec(ShortCut).Lo - $30 + Ord('0'));
    $41..$5A: Name := Chr(WordRec(ShortCut).Lo - $41 + Ord('A'));
    $60..$69: Name := Chr(WordRec(ShortCut).Lo - $60 + Ord('0'));
    $70..$87: Name := 'F' + IntToStr(WordRec(ShortCut).Lo - $6F);
    else
      Name := scGetSpecialKeyName(ShortCut);
  end;
  if Name <> '' then
  begin
    Result := '';
    if ShortCut and scCtrl <> 0 then
      Result := Result + SCKeyCaps[sckcCtrl];
    if ShortCut and scShift <> 0 then
      Result := Result + SCKeyCaps[sckcShift];
    if ShortCut and scAlt <> 0 then
      Result := Result + SCKeyCaps[sckcAlt];
    Result := Result + Name;
  end else
    Result := '';
end;

function scShiftStateToKeys(Shift: TShiftState): Word;
begin
  Result := 0;
  if ssShift in Shift then
    Result := Result or SC_SHIFT;
  if ssCtrl in Shift then
    Result := Result or SC_CONTROL;
  if ssLeft in Shift then
    Result := Result or SC_LBUTTON;
  if ssRight in Shift then
    Result := Result or SC_RBUTTON;
  if ssMiddle in Shift then
    Result := Result or SC_MBUTTON;
  if ssAlt in Shift then
    Result := Result or SC_MENU;
end;

function scKeysToShiftState(Keys: Word): TShiftState;
begin
  Result := [];
  if Keys and MK_SHIFT <> 0 then
    Include(Result, ssShift);
  if Keys and MK_CONTROL <> 0 then
    Include(Result, ssCtrl);
  if Keys and MK_LBUTTON <> 0 then
    Include(Result, ssLeft);
  if Keys and MK_RBUTTON <> 0 then
    Include(Result, ssRight);
  if Keys and MK_MBUTTON <> 0 then
    Include(Result, ssMiddle);
  if GetKeyState(VK_MENU) < 0 then
    Include(Result, ssAlt);
end;

function scKeyDataToShiftState(KeyData: Longint): TShiftState;
const
  AltMask = $20000000;
begin
  Result := [];
  if GetKeyState(VK_SHIFT) < 0 then
    Include(Result, ssShift);
  if GetKeyState(VK_CONTROL) < 0 then
    Include(Result, ssCtrl);
  if KeyData and AltMask <> 0 then
    Include(Result, ssAlt);
end;

function scKeyboardStateToShiftState(const KeyboardState: TKeyboardState): TShiftState;
begin
  Result := [];
  if KeyboardState[VK_SHIFT] and $80 <> 0 then
    Include(Result, ssShift);
  if KeyboardState[VK_CONTROL] and $80 <> 0 then
    Include(Result, ssCtrl);
  if KeyboardState[VK_MENU] and $80 <> 0 then
    Include(Result, ssAlt);
  if KeyboardState[VK_LBUTTON] and $80 <> 0 then
    Include(Result, ssLeft);
  if KeyboardState[VK_RBUTTON] and $80 <> 0 then
    Include(Result, ssRight);
  if KeyboardState[VK_MBUTTON] and $80 <> 0 then
    Include(Result, ssMiddle);
end;

{ THotkeyEdit }

procedure THotkeyEdit.UpdateHotKey(Value: TShortCut);
begin
  if FHotkeyUpdate = 0 then
  begin
    Inc(FHotkeyUpdate);
    try
      SetHotKey(Value);
    finally
      Dec(FHotkeyUpdate);
    end;
  end;
end;

function THotkeyEdit.GetHotKey: TShortCut;
var
  Shift: TShiftState;
begin
  Shift := [];

  if schkShift in FHotModifiers then
    Include(Shift, ssShift);
  if schkCtrl in FHotModifiers then
    Include(Shift, ssCtrl);
  if schkAlt in FHotModifiers then
    Include(Shift, ssAlt);

  if (Shift = []) then // and not (schcNone in FInvalidKeys) then
  begin
    if schkShift in FModifiers then
      Include(Shift, ssShift);
    if schkCtrl in FModifiers then
      Include(Shift, ssCtrl);
    if schkAlt in FModifiers then
      Include(Shift, ssAlt);
  end;

  Result := scShortCut(FHotKey, Shift);
end;

procedure THotkeyEdit.SetHotKey(Value: TShortCut);
begin
  ShortCutToHotKey(Value);
  UpdateHotText;

  //if Assigned(FOnHotkeyChange) and (Sc <> GetHotKey) then
  //  FOnHotkeyChange(Self);
end;

procedure THotkeyEdit.ShortCutToHotKey(Value: TShortCut);
begin
  if (Value = VK_BACK) or (Value = VK_DELETE) then
    Value := 0;

  FHotKey := Value and not (scShift + scCtrl + scAlt);;
  FHotModifiers := [];

  if Value and scShift <> 0 then
    Include(FHotModifiers, schkShift);
  if Value and scCtrl <> 0 then
    Include(FHotModifiers, schkCtrl);
  if Value and scAlt <> 0 then
    Include(FHotModifiers, schkAlt);

  if (schcShiftCtrlAlt in FInvalidKeys) and (FHotModifiers = [schkShift, schkCtrl, schkAlt]) then
    FHotModifiers := [];

  if (schcCtrlAlt in FInvalidKeys) and (FHotModifiers = [schkCtrl, schkAlt]) then
    FHotModifiers := [];

  if (schcShiftAlt in FInvalidKeys) and (FHotModifiers = [schkShift, schkAlt]) then
    FHotModifiers := [];

  if (schcShiftCtrl in FInvalidKeys) and (FHotModifiers = [schkShift, schkCtrl]) then
    FHotModifiers := [];

  if (schcAlt in FInvalidKeys) and (FHotModifiers = [schkAlt]) then
    FHotModifiers := [];

  if (schcCtrl in FInvalidKeys) and (FHotModifiers = [schkCtrl]) then
    FHotModifiers := [];

  if (schcShift in FInvalidKeys) and (FHotModifiers = [schkShift]) then
    FHotModifiers := [];

  if (schcNone in FInvalidKeys) and (FHotModifiers = []) then
    FHotModifiers := [];

  if (FHotModifiers = []) then // and not (schcNone in FInvalidKeys) then
  begin
    if schkShift in FModifiers then
      Include(FHotModifiers, schkShift);
    if schkCtrl in FModifiers then
      Include(FHotModifiers, schkCtrl);
    if schkAlt in FModifiers then
      Include(FHotModifiers, schkAlt);
  end;
end;

procedure THotkeyEdit.UpdateHotText;
var
  S: String;
begin
  Inc(FInternalUpdate);
  try
    if FHotKey <> 0 then
      S := ShortCutToText(GetHotKey)
    else
    begin
      if schkCtrl in FHotModifiers then
        S := S + scskCtrl;
      if schkShift in FHotModifiers then
        S := S + scskShift;
      if schkAlt in FHotModifiers then
        S := S + scskAlt;
    end;

    RealSetText(S);
  finally
    Dec(FInternalUpdate);
  end;
end;

procedure THotkeyEdit.KeyDown(var Key: Word; Shift: TShiftState);
var
  K: Word;
begin
  K := Key;
  Key := 0;
  FHotModifiers := [];

  if K in [VK_SHIFT, VK_CONTROL, VK_MENU] then
    K := 0;

  UpdateHotKey(scShortCut(K, Shift));
end;

procedure THotkeyEdit.RealSetText(const AValue: TCaption);
begin
  if FInternalUpdate > 0 then
  begin
    inherited RealSetText(AValue);
    SetCaretPos(TPoint.Create(Length(Text), 0));
  end;
end;

procedure THotkeyEdit.KeyPress(var Key: char);
begin
  Key := #0;
  inherited KeyPress(Key);
end;

procedure THotkeyEdit.KeyUp(var Key: Word; Shift: TShiftState);
begin
  Key := 0;
  if FHotKey = 0 then
  begin
    FHotModifiers := [];
    UpdateHotText;
  end;
end;

end.
